from thinc.v2v import Affine
from thinc.neural.optimizers import SGD
import numpy
try:
    import torch.nn
    from thinc.extra.wrappers import PyTorchWrapper
except ImportError:
    PyTorchWrapper = None

def check_learns_zero_output(model, sgd, X, Y):
    '''Check we can learn to output a zero vector'''
    Yh, get_dX = model.begin_update(X)
    dYh = (Yh-Y) / Yh.shape[0]
    dX = get_dX(dYh, sgd=sgd)
    prev = numpy.abs(Yh.sum())
    for i in range(100):
        Yh, get_dX = model.begin_update(X)
        total = numpy.abs(Yh.sum())
        dX = get_dX(Yh-Y, sgd=sgd)
        assert total < prev
        prev = total


def test_unwrapped(nN=2, nI=3, nO=4):
    if PyTorchWrapper is None:
        return
    model = Affine(nO, nI)
    X = numpy.zeros((nN, nI), dtype='f')
    X += numpy.random.uniform(size=X.size).reshape(X.shape)
    sgd = SGD(model.ops, 0.001)
    Y = numpy.zeros((nN, nO), dtype='f')
    check_learns_zero_output(model, sgd, X, Y)

def test_wrapper(nN=2, nI=3, nO=4):
    if PyTorchWrapper is None:
        return
    model = PyTorchWrapper(torch.nn.Linear(nI, nO))
    sgd = SGD(model.ops, 0.001)
    X = numpy.zeros((nN, nI), dtype='f')
    X += numpy.random.uniform(size=X.size).reshape(X.shape)
    Y = numpy.zeros((nN, nO), dtype='f')
    Yh, get_dX = model.begin_update(X)
    assert Yh.shape == (nN, nO)
    dYh = (Yh-Y) / Yh.shape[0]
    dX = get_dX(dYh, sgd=sgd)
    assert dX.shape == (nN, nI)
    check_learns_zero_output(model, sgd, X, Y)



cdef void simple_gemm(float* output, int o0, int o1,
                     const float* A, int a0, int a1,
                     const float* B, int b0, int b1,
                     int trans1, int trans2) nogil
 
#cdef void simple_ger(float* output, int o0, int o1,
#                     const float* A, int a0,
#                     const float* B, int b0) nogil
#
#cdef void scale(float* output, int o0, float scale) nogil

cdef void simple_axpy(float* output, int o0,
    const float* A, float scale) nogil
 



# cython: infer_types=True
# cython: cdivision=True
from .typedefs cimport weight_t


cdef extern from "cblas.h":
    enum CBLAS_ORDER:     CblasRowMajor, CblasColMajor
    enum CBLAS_TRANSPOSE: CblasNoTrans, CblasTrans, CblasConjTrans
    enum CBLAS_UPLO:      CblasUpper, CblasLower
    enum CBLAS_DIAG:      CblasNonUnit, CblasUnit
    enum CBLAS_SIDE:      CblasLeft, CblasRight

    # BLAS level 1 routines

    #void cblas_sswap(int M, float  *x, int incX, float  *y, int incY) nogil
    #void cblas_sscal(int N, float  alpha, float  *x, int incX) nogil
    #void cblas_scopy(int N, float  *x, int incX, float  *y, int incY) nogil
    void cblas_saxpy(int N, float  alpha, float  *x, int incX, float  *y, int incY ) nogil
    #void cblas_dswap(int M, double  *x, int incX, double  *y, int incY) nogil
    #void cblas_dscal(int N, double  alpha, double *x, int incX) nogil
    #void cblas_dcopy(int N, double *x, int incX, double  *y, int incY) nogil
    #void cblas_daxpy(int N, double  alpha, double *x, int incX, double *y, int incY ) nogil

    #float cblas_sdot(int N, float  *x, int incX, float *y, int incY ) nogil
    #float cblas_snrm2(int N, float  *x, int incX) nogil
    #float cblas_sasum(int N, float  *x, int incX) nogil
    #int cblas_isamax(int N, float  *x, int incX) nogil
    #double cblas_ddot(int N, double  *x, int incX, double  *y, int incY ) nogil
    #double cblas_dnrm2(int N, double  *x, int incX) nogil
    #double cblas_dasum(int N, double  *x, int incX) nogil
    #int cblas_idamax(int N, double  *x, int incX) nogil

    # BLAS level 2 routines
    #void cblas_sgemv(CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA, int M, int N,
    #                float  alpha, float *A, int lda, float *x, int incX,
    #                             float beta, float *y, int incY) nogil

    #void cblas_sger(CBLAS_ORDER Order, int M, int N, float  alpha, float  *x,
    #                            int incX, float  *y, int incY, float  *A, int lda) nogil

    #void cblas_dgemv(CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA, int M, int N,
    #                 double  alpha, double  *A, int lda, double  *x, int incX,
    #                 double  beta, double  *y, int incY) nogil

    #void cblas_dger(CBLAS_ORDER Order, int M, int N, double  alpha, double  *x,
    #                int incX, double  *y, int incY, double  *A, int lda) nogil

    # BLAS level 3 routines
    void cblas_sgemm(CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA,
                                 CBLAS_TRANSPOSE TransB, int M, int N, int K,
                                 float  alpha, float  *A, int lda, float  *B, int ldb,
                                 float  beta, float  *C, int ldc) nogil
  
    #void cblas_dgemm(CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA,
    #                             CBLAS_TRANSPOSE TransB, int M, int N, int K,
    #                             double  alpha, double  *A, int lda, double  *B, int ldb,
    #                             double  beta, double  *C, int ldc) nogil


cdef void simple_gemm(float* output, int o0, int o1,
                     const float* A, int a0, int a1,
                     const float* B, int b0, int b1,
                     int trans1, int trans2) nogil:
    cdef float alpha = 1.0
    cdef float beta = 1.0
    if not trans1 and not trans2:
        cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                    o0, o1, b0, alpha, A, a1, B, b1, beta, output, o1)
    elif not trans1 and trans2:
        cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasTrans,
                    o0, o1, b1, alpha, A, a1, B, b1, beta, output, o1)
    elif trans1 and trans2:
        cblas_sgemm(CblasRowMajor, CblasTrans, CblasTrans,
                    o0, o1, b1, alpha, A, a1, B, b1, beta, output, o1)
    elif trans1 and not trans2:
        cblas_sgemm(CblasRowMajor, CblasTrans, CblasNoTrans,
                    o0, o1, b0,
                    alpha, A, a1, B, b1,
                    beta, output, o1)
    else:
        with gil:
            dims = '(%d, %d) = (%d, %d) @ (%d, %d)' % (o0, o1, a0, a1, b0, b1)
            raise ValueError("Invalid dimensions for GEMM: %s" % dims)
 

cdef void simple_ger(float* output, int o0, int o1,
                     const float* A, int a0,
                     const float* B, int b0) nogil:
    cdef float alpha = 1.
    pass
    #cblas_sger(CblasRowMajor, o0, o1, alpha, A, 1, B, 1,
    #    output, o1)


#cdef void scale(float* output, int o0, float scale) nogil:
#    cblas_sscal(o0, scale, output, 1)


cdef void simple_axpy(float* output, int o0,
        const float* A, float scale) nogil:

    cblas_saxpy(o0, 1., A, 1, output, 1)



import pytest
from ..neural.util import get_ops
from ..neural.ops import NumpyOps, CupyOps


def test_get_ops():
    Ops = get_ops('numpy')
    Ops is NumpyOps
    Ops = get_ops('cpu')
    assert Ops is NumpyOps
    Ops = get_ops('cupy')
    assert Ops is CupyOps
    Ops = get_ops('gpu')
    assert Ops is CupyOps
    with pytest.raises(ValueError):
        Ops = get_ops('blah')






from .model import Model
from ... import describe
from ...describe import Dimension, Synapses, Biases, Gradient
from ...api import wrap, layerize
from .._lsuv import svd_orthonormal
from ..util import copy_array


def BiLSTM(nO, nI):
    '''Create a bidirectional LSTM layer. Args: number out, number in'''
    return Bidirectional(LSTM(nO//2, nI), LSTM(nO//2, nI))


def LSTM(nO, nI):
    '''Create an LSTM layer. Args: number out, number in'''
    weights = LSTM_weights(nO, nI)
    gates = LSTM_gates(weights.ops)
    return Recurrent(RNN_step(weights, gates))


def Bidirectional(l2r, r2l):
    '''Stitch two RNN models into a bidirectional layer.'''
    nO = l2r.nO
    def birnn_fwd(Xs, drop=0.):
        l2r_Zs, bp_l2r_Zs = l2r.begin_update(Xs, drop=drop) 
        r2l_Zs, bp_r2l_Zs = r2l.begin_update([l2r.ops.xp.ascontiguousarray(X[::-1])
                                              for X in Xs]) 
        def birnn_bwd(dZs, sgd=None):
            d_l2r_Zs = []
            d_r2l_Zs = []
            for dZ in dZs:
                l2r_fwd = dZ[:, :nO]
                r2l_fwd = dZ[:, nO:]
                d_l2r_Zs.append(l2r.ops.xp.ascontiguousarray(l2r_fwd))
                d_r2l_Zs.append(l2r.ops.xp.ascontiguousarray(r2l_fwd[::-1]))
            dXs_l2r = bp_l2r_Zs(d_l2r_Zs, sgd=sgd)
            dXs_r2l = bp_r2l_Zs(d_r2l_Zs, sgd=sgd)
            dXs = [dXf+dXb[::-1] for dXf, dXb in zip(dXs_l2r, dXs_r2l)]
            return dXs
        Zs = [l2r.ops.xp.hstack((Zf, Zb[::-1])) for Zf, Zb in zip(l2r_Zs, r2l_Zs)]
        return Zs, birnn_bwd
    return wrap(birnn_fwd, l2r, r2l)


def Recurrent(step_model):
    '''Apply a stepwise model over a sequence, maintaining state. For RNNs'''
    ops = step_model.ops
    def recurrent_fwd(seqs, drop=0.):
        lengths = [len(X) for X in seqs]
        X, size_at_t, unpad = ops.square_sequences(seqs)
        Y = ops.allocate((X.shape[0], X.shape[1], step_model.nO))
        cell_drop = ops.get_dropout_mask((len(seqs), step_model.nO), 0.)
        hidden_drop = ops.get_dropout_mask((len(seqs), step_model.nO), 0.)
        out_drop = ops.get_dropout_mask((len(seqs), step_model.nO), 0.)
        backprops = [None] * max(lengths)
        state = step_model.weights.get_initial_state(len(seqs))
        for t in range(max(lengths)):
            state = list(state)
            size = size_at_t[t]
            Xt = X[t, :size]
            state[0] = state[0][:size]
            state[1] = state[1][:size]
            if cell_drop is not None:
                state[0] *= cell_drop
            if hidden_drop is not None:
                state[1] *= hidden_drop
            inputs = (state, Xt)
            (state, Y[t, :size]), backprops[t] = step_model.begin_update(inputs)
            if out_drop is not None:
                Y[t, :size] *= out_drop
        outputs = unpad(Y)

        def recurrent_bwd(d_outputs, sgd=None):
            dY, size_at_t, unpad = step_model.ops.square_sequences(d_outputs)
            d_state = [step_model.ops.allocate((dY.shape[1], step_model.nO)),
                       step_model.ops.allocate((dY.shape[1], step_model.nO))]
            updates = {}
            def gather_updates(weights, gradient, key=None):
                updates[key] = (weights, gradient)

            dX = step_model.ops.allocate((dY.shape[0], dY.shape[1], step_model.weights.nI))
            for t in range(max(lengths)-1, -1, -1):
                if out_drop is not None:
                    dY[t] *= out_drop
                d_state_t, dXt = backprops[t]((d_state, dY[t]),
                                            sgd=gather_updates)
                d_state[0][:d_state_t[0].shape[0]] = d_state_t[0]
                d_state[1][:d_state_t[1].shape[0]] = d_state_t[1]
                dX[t, :dXt.shape[0]] = dXt
                if cell_drop is not None:
                    d_state[0] *= cell_drop
                if hidden_drop is not None:
                    d_state[1] *= hidden_drop
            d_cell, d_hidden = d_state
            step_model.weights.d_initial_cells += d_cell.sum(axis=0)
            step_model.weights.d_initial_hiddens += d_hidden.sum(axis=0)
            if sgd is not None:
                for key, (weights, gradient) in updates.items():
                    sgd(weights, gradient, key=key)
            return unpad(dX)
        return outputs, recurrent_bwd
    model = wrap(recurrent_fwd, step_model)
    model.nO = step_model.nO
    return model


def RNN_step(weights, gates):
    '''Create a step model for an RNN, given weights and gates functions.'''
    def rnn_step_fwd(prevstate_inputs, drop=0.):
        prevstate, inputs = prevstate_inputs
        cell_tm1, hidden_tm1 = prevstate

        acts, bp_acts = weights.begin_update((inputs, hidden_tm1), drop=drop)
        (cells, hiddens), bp_gates = gates.begin_update((acts, cell_tm1), drop=drop)
        size = cell_tm1.shape[0]

        def rnn_step_bwd(d_state_d_hiddens, sgd=None):
            (d_cells, d_hiddens), d_hiddens = d_state_d_hiddens
            d_acts, d_cell_tm1 = bp_gates((d_cells, d_hiddens), sgd=sgd)
            d_inputs, d_hidden_tm1 = bp_acts(d_acts, sgd=sgd)
            return (d_cell_tm1, d_hidden_tm1), d_inputs
        return ((cells, hiddens), hiddens), rnn_step_bwd
    model = wrap(rnn_step_fwd, weights, gates)
    model.nO = weights.nO
    model.nI = weights.nI
    model.weights = weights
    model.gates = gates
    return model


def LSTM_gates(ops):
    def lstm_gates_fwd(acts_prev_cells, drop=0.):
        acts, prev_cells = acts_prev_cells
        new_cells = ops.allocate(prev_cells.shape)
        new_hiddens = ops.allocate(prev_cells.shape)
        ops.lstm(new_hiddens, new_cells,
            acts, prev_cells)
        state = (new_cells, new_hiddens)
        size = new_cells.shape[0]

        def lstm_gates_bwd(d_state, sgd=None):
            d_cells, d_hiddens = d_state
            d_cells = d_cells[:size]
            d_hiddens = d_hiddens[:size]
            d_acts = [ops.allocate(act.shape) for act in acts]
            d_prev_cells = ops.allocate(prev_cells.shape)
            ops.backprop_lstm(d_cells, d_prev_cells, d_acts,
                    d_hiddens, acts, new_cells, prev_cells)
            return d_acts, d_prev_cells
        return state, lstm_gates_bwd
    return layerize(lstm_gates_fwd)


def _uniform_init(lo, hi):
    def wrapped(W, ops):
        copy_array(W, ops.xp.random.uniform(lo, hi, W.shape))
    return wrapped


@describe.attributes(
    nO=Dimension("Output size"),
    nI=Dimension("Input size"),
    W=Synapses("Weights matrix",
        lambda obj: (obj.nO * 4, obj.nI + obj.nO),
        lambda W, ops: copy_array(W, svd_orthonormal(W.shape))),
    b=Biases("Bias vector",
        lambda obj: (obj.nO * 4,)),
    forget_bias=Biases("Bias for forget gates",
        lambda obj: (obj.nO,),
        lambda b, ops: copy_array(b, ops.xp.ones(b.shape, dtype=b.dtype))),
    d_W=Gradient("W"),
    d_b=Gradient("b"),
    d_forget_bias=Gradient("forget_bias"),
    initial_hiddens=Biases("Initial hiddens", lambda obj: (obj.nO,),
        _uniform_init(-0.1, 0.1)),
    initial_cells=Biases("Initial cells", lambda obj: (obj.nO,),
        _uniform_init(-0.1, 0.1)),
    d_initial_hiddens=Gradient("initial_hiddens"),
    d_initial_cells=Gradient("initial_cells")
)
class LSTM_weights(Model):
    def __init__(self, nO, nI):
        Model.__init__(self)
        self.nO = nO
        self.nI = nI

    def begin_update(self, inputs_hidden, drop=0.):
        inputs, hidden = inputs_hidden
        assert inputs.dtype == 'float32'
        X = self.ops.xp.hstack([inputs, hidden])
        acts = self.ops.gemm(X, self.W, trans2=True) + self.b
        acts = self._split_activations(acts)
        acts[0] += self.forget_bias
        def bwd_lstm_weights(d_acts, sgd=None):
            self.d_forget_bias += d_acts[0].sum(axis=0)
            d_acts = self._merge_activations(d_acts)
            dX = self.ops.gemm(d_acts, self.W)
            self.d_W += self.ops.gemm(d_acts, X, trans1=True)
            self.d_b += d_acts.sum(axis=0)
            d_input = dX[:, :self.nI]
            d_hidden = dX[:, self.nI:]
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return d_input, d_hidden
        return acts, bwd_lstm_weights

    def get_initial_state(self, n):
        initial_cells = self.ops.allocate((n, self.nO))
        initial_hiddens = self.ops.allocate((n, self.nO))
        initial_cells += self.initial_cells
        initial_hiddens += self.initial_hiddens
        return (initial_cells, initial_hiddens)

    def _split_activations(self, acts):
        acts = acts.reshape((acts.shape[0], 4, self.nO))
        acts = self.ops.xp.ascontiguousarray(acts.transpose((1, 0, 2)))
        return [acts[0], acts[1], acts[2], acts[3]]

    def _merge_activations(self, act_pieces):
        return self.ops.xp.hstack(act_pieces)



from __future__ import print_function
import pytest

from ...neural.vec2vec import Model, ReLu, Softmax
from ...neural._classes.feed_forward import FeedForward
from ...neural._classes.batchnorm import BatchNorm
from ...neural._classes.embed import Embed
from ...neural._classes.convolution import ExtractWindow
from ...neural.ops import NumpyOps
from ...api import layerize, clone, chain, with_flatten
from ...loss import categorical_crossentropy

from ...extra import datasets


@pytest.fixture(scope='module')
def ancora():
    train_data, check_data, nr_class = datasets.ancora_pos_tags()
    train_X, train_y = zip(*train_data)
    dev_X, dev_y = zip(*check_data)
    return (train_X[:100], train_y[:100]), (dev_X, dev_y)

@pytest.fixture(scope='module')
def train_X(ancora):
    (train_X, train_y), (dev_X, dev_y) = ancora
    return train_X


@pytest.fixture(scope='module')
def train_y(ancora):
    (train_X, train_y), (dev_X, dev_y) = ancora
    return train_y


@pytest.fixture(scope='module')
def dev_X(ancora):
    (train_X, train_y), (dev_X, dev_y) = ancora
    return dev_X


@pytest.fixture(scope='module')
def dev_y(ancora):
    (train_X, train_y), (dev_X, dev_y) = ancora
    return dev_y


def create_embed_relu_relu_softmax(depth, width, vector_length):
    with Model.define_operators({'>>': chain}):
        model = (
                with_flatten(
                  Embed(width, vector_length)
                  >> ExtractWindow(nW=1)
                  >> ReLu(width)
                  >> ReLu(width)
                  >> Softmax(20)
                ))
    return model


@pytest.fixture(params=[create_embed_relu_relu_softmax])
def create_model(request):
    return request.param


@pytest.mark.slow
@pytest.mark.parametrize(('depth', 'width', 'vector_width', 'nb_epoch'),
        [(2, 32, 16, 5)])
def test_small_end_to_end(depth, width, vector_width, nb_epoch,
        create_model,
        train_X, train_y, dev_X, dev_y):
    model = create_model(depth, width, vector_width)
    assert isinstance(model, Model)
    losses = []
    with model.begin_training(train_X, train_y) as (trainer, optimizer):
        trainer.nb_epoch = 10
        trainer.batch_size = 8
        for X, y in trainer.iterate(train_X, train_y):
            yh, backprop = model.begin_update(X, drop=trainer.dropout)
            d_loss = []
            loss = []
            for i in range(len(yh)):
                dl, l = categorical_crossentropy(yh[i], y[i])
                d_loss.append(dl)
                loss.append(l)
            backprop(d_loss, optimizer)
            losses.append(sum(loss))
    assert losses[-1] < losses[0]



import numpy

from ...import describe
from ...describe import Dimension, Synapses, Gradient
from .._lsuv import LSUVinit
from ..ops import NumpyOps
from .model import Model
from ...extra.load_nlp import get_vectors

try:
    import cupy
except ImportError:
    cupy = None


#@describe.on_data(LSUVinit)
@describe.attributes(
        nM=Dimension("Vector dimensions"),
        nO=Dimension("Size of output"),
        W=Synapses(
            "A projection matrix, to change vector dimensionality",
            lambda obj: (obj.nO, obj.nM),
            lambda W, ops: ops.xavier_uniform_init(W)),
        d_W=Gradient("W"),
)
class StaticVectors(Model):
    '''Load a static embedding table, and learn a linear projection from it.

    Out-of-vocabulary items are modded into the table, receiving an arbitrary
    vector (but the same word will always receive the same vector).
    '''
    name = 'static-vectors'
    def __init__(self, lang, nO, drop_factor=0.0, column=0):
        Model.__init__(self)
        self.column = column
        self.nO = nO
        # This doesn't seem the cleverest solution,
        # but it ensures multiple models load the
        # same copy of spaCy if they're deserialised.
        self.lang = lang
        vectors = self.get_vectors()
        self.nM = vectors.shape[1]
        self.drop_factor = drop_factor
        self.column = column
        if self.nM == 0:
            raise ValueError(
                "Cannot create vectors table with dimension 0.\n"
                "If you're using pre-trained vectors, are the vectors loaded?")
        self.nV = vectors.shape[0]

    def get_vectors(self):
        return get_vectors(self.ops, self.lang)

    def begin_update(self, ids, drop=0.):
        if ids.ndim >= 2:
            ids = self.ops.xp.ascontiguousarray(ids[:, self.column])
        vector_table = self.get_vectors()
        vectors = vector_table[ids * (ids < vector_table.shape[0])]
        vectors = self.ops.xp.ascontiguousarray(vectors)
        assert vectors.shape[0] == ids.shape[0]
        def finish_update(gradients, sgd=None):
            if mask is not None:
                gradients *= mask
            self.d_W += self.ops.gemm(gradients, vectors, trans1=True)
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return None
        dotted = self.ops.gemm(vectors, self.W, trans2=True)
        mask = self.ops.get_dropout_mask((dotted.shape[1],), drop)
        if mask is not None:
            dotted *= mask
        return dotted, finish_update



import pytest
import time
from numpy.testing import assert_allclose
import numpy.random
from timeit import default_timer as timer
from ...neural._classes.maxout import Maxout

from ...neural._fast_maxout_cnn import MaxoutWindowEncoder

numpy.random.seed(0)

def test_create():
    mwe = MaxoutWindowEncoder(16, 4)


def test_fwd_runs():
    mwe = MaxoutWindowEncoder(32, 4)
    X = mwe.ops.allocate((5, 32), dtype='f')
    y = mwe([X])[0]
    assert y.shape == X.shape
    assert y.sum() == 0.
    y += 2
    z = mwe([y])[0]
    assert z.sum() != 0.


def test_bwd_runs():
    mwe = MaxoutWindowEncoder(32, 4)
    X = mwe.ops.allocate((5, 32), dtype='f')
    dy = mwe.ops.allocate((5, 32), dtype='f')
    y, bp_y = mwe.begin_update([X])
    dX = bp_y([dy])


def baseline_mwe(nO, nP, depth):
    from thinc.neural._classes.model import Model
    from thinc.neural._classes.resnet import Residual
    from thinc.neural._classes.convolution import ExtractWindow
    from thinc.neural._classes.layernorm import LayerNorm
    from thinc.api import chain, clone, with_flatten
    maxout = Maxout(nO, nO*3, pieces=nP)
    normalize = LayerNorm(maxout)
    with Model.define_operators({'>>': chain, '**': clone}):
        model = Residual(ExtractWindow(nW=1) >> normalize)
        model = with_flatten(chain(*([model]*depth)))
    model.maxout = maxout
    model.normalize = normalize
    return model


def test_fwd_correctness(nr_row=20, nr_dim=5, nr_piece=3):
    base = baseline_mwe(nr_dim, 3, 4)
    fast = MaxoutWindowEncoder(nr_dim, 4)
    fast.maxout.W[:] = base.maxout.W
    fast.normalize.G[:] = base.normalize.G
    Xs = [fast.ops.normal_init(fast.ops.allocate((nr_row, nr_dim)), nr_dim)
          for _ in range(10)]
    Ys_new = fast(Xs)
    # Because of the flattening, this isn't correct if just base(Xs)
    Ys_old = [base([X])[0] for X in Xs]
    for Y1, Y2 in zip(Ys_new, Ys_old):
        assert_allclose(Y1, Y2, rtol=0.0001, atol=0.0001)

@pytest.mark.xfail
def test_bwd_correctness(nr_row=2, nr_dim=2, nr_piece=3):
    base = baseline_mwe(nr_dim, 3, 2)
    fast = MaxoutWindowEncoder(nr_dim, 2)
    fast.maxout.W[:] = base.maxout.W
    fast.normalize.G[:] = base.normalize.G
    Xs = [fast.ops.normal_init(fast.ops.allocate((nr_row, nr_dim)), nr_dim)
          for _ in range(3)]
    Ys, bp_Ys = fast.begin_update(Xs)
    dXs_new = bp_Ys(Xs)
    dXs_old = []
    for X in Xs:
        Y, bp_Y = base.begin_update([X])
        dXs_old.append(bp_Y([X])[0])
    for dX1, dX2 in zip(dXs_new, dXs_old):
        assert_allclose(dX1, dX2, rtol=1e-2, atol=1e-2)


def test_fwd_speed(nr_row=100, nr_dim=128, nr_piece=3):
    mwe = MaxoutWindowEncoder(nr_dim, 4)
    Xs = [mwe.ops.allocate((nr_row, nr_dim)) for _ in range(100)]
    start = timer()
    ys = mwe(Xs)
    end = timer()
    print('Fwd Fast?', end, start, end-start)
    mwe = baseline_mwe(nr_dim, nr_piece, 4)
    start = timer()
    y = mwe(Xs)
    end = timer()
    print('Fwd Slow?', end, start, end-start)


def test_bwd_speed(nr_row=30, nr_dim=128, nr_piece=3):
    mwe = MaxoutWindowEncoder(nr_dim, 4)
    Xs = [mwe.ops.normal_init(mwe.ops.allocate((nr_row, nr_dim)), nr_dim)
          for _ in range(100)]
    start = timer()
    ys, bp_ys = mwe.begin_update(Xs)
    dx = bp_ys(Xs)
    end = timer()
    print('Fast?', end, start, '%.4f' % (end-start))

    base = baseline_mwe(nr_dim, nr_piece, 4)
    start = timer()
    ys, bp_ys = base.begin_update(Xs)
    dx = bp_ys(Xs)
    end = timer()
    print('Slow?', end, start, '%.4f' % (end-start))


